@@ -2,9 +2,8 @@
 
 Provides test clients for in-memory testing without subprocess/network overhead.
 """
-from django_bolt.testing.client import TestClient, AsyncTestClient
+from django_bolt.testing.client import TestClient
 
 __all__ = [
     "TestClient",
-    "AsyncTestClient",
 ]
